import numpy as np
import torch
from torch import Tensor


def batch_mahalanobis_sq(x: Tensor, mu: Tensor, sigma_inv: Tensor) -> Tensor:
    """
    Allows calling `mahalanobis_sq` on a batch of input tensors, for multiple
    distributions, look at `mahalanobis_multi` instead.
    Params
    ======
        x: The input tensors of size (b, h * w, c)
        mu: The mean tensors of size (h * w, c)
        sigma_inv: The inverted covariance matrices of size (h * w, c, c)
    Returns
    =======
        dist: Distance tensor of size (b, h * w, 1)
    """
    result = mahalanobis_sq(x, mu, sigma_inv)
    return torch.diagonal(result, dim1=1,
                          dim2=2).squeeze(1).unsqueeze(-1)  # (b, h * w, 1)


def mahalanobis_sq(x: Tensor, mu: Tensor, sigma_inv: Tensor) -> Tensor:
    """
    The squared mahalanobis distance using PyTorch
    Params
    ======
        x: The input tensors of size (h * w, c)
        mu: The mean tensors of size (h * w, c)
        sigma_inv: The inverted covariance matrices of size (h * w, c, c)
    Returns
    =======
        dist: Distance tensor of size (h * w, 1)
    """
    delta = x - mu  # (h * w, c)
    temp = torch.matmul(sigma_inv, delta.unsqueeze(-1))  # (h * w, c, 1)
    dist = torch.matmul(delta.unsqueeze(1), temp)  # (h * w, 1, 1)
    return dist.squeeze(1)  # (h * w, 1)


def mahalanobis_multi(x, means, sigmas):
    """
    Computes the mahalanobis distance with regard to multiple distributions
    Params
    ======
        x: Tensor - size (h * w, c)
        means: Tensor - size (n, c)
        sigmas: Tensor - size (n, c, c)
    Returns
    =======
        distances: Tensor - size (n, h * w, 1)
    """
    n_mixtures, embeddings_size = means.shape
    n_patches = x.size(0)
    distances = torch.zeros((n_mixtures, n_patches, 1))
    for i in range(n_mixtures):
        distances[i, :, :] = mahalanobis_sq(x, means[i], sigmas[i])
    return distances



from .utils import (
    embeddings_concat,
    mean_smoothing,
    compute_roc_score,
    compute_pro_score
)
from .regions import (
    propose_region,
    propose_regions,
    propose_regions_cv2,
    IoU,
    floating_IoU,
)

__all__ = [
    "embeddings_concat",
    "mean_smoothing",
    "compute_roc_score",
    "compute_pro_score",
    "propose_region",
    "propose_regions",
    "propose_regions_cv2",
    "IoU",
    "floating_IoU",
]



from typing import Union, Tuple

import torch
from torch import Tensor, device as Device

from src.anoseg.models.PaDiM.utils import embeddings_concat
from src.anoseg.models.PaDiM.backbone import ResNet18, ResNet50, WideResNet50


class PaDiMBase:
    """The embedding backbone shared by PaDiM and PaDiMSVDD
    """

    def __init__(self, num_embeddings: int, device: Union[str, Device],
                 backbone: str, size=None):
        self.device = device
        self.num_embeddings = num_embeddings

        if size is not None:
            self._init_backbone_with_size(backbone, size)
        else:
            self._init_backbone(backbone)

        self.embedding_ids = torch.randperm(
            self.max_embeddings_size)[:self.num_embeddings].to(self.device)

    def _get_backbone(self):
        if isinstance(self.model, ResNet18):
            backbone = "resnet18"
        elif isinstance(self.model, ResNet50):
            backbone = "resnet50"
        elif isinstance(self.model, WideResNet50):
            backbone = "wide_resnet50"
        else:
            raise NotImplementedError()

        return backbone

    def _init_backbone_with_size(self, backbone: str, size: Tuple[int, int]) -> None:
        self._init_backbone(backbone)
        empty_batch = torch.zeros((1, 3) + size, device=self.device)
        feature_1, _, _ = self.model(empty_batch)
        _, _, w, h = feature_1.shape
        self.num_patches = w * h
        self.model.num_patches = w * h

    def _init_backbone(self, backbone: str) -> None:
        if backbone == "resnet18":
            self.model = ResNet18().to(self.device)
        elif backbone == "resnet50":
            self.model = ResNet50().to(self.device)
        elif backbone == "wide_resnet50":
            self.model = WideResNet50().to(self.device)
        else:
            raise Exception(f"unknown backbone {backbone}, "
                            "choose one of ['resnet18', 'resnet50', 'wide_resnet50']")

        self.num_patches = self.model.num_patches
        self.max_embeddings_size = self.model.embeddings_size

    def _embed_batch(self, imgs: Tensor, with_grad: bool = False) -> Tensor:
        self.model.eval()
        with torch.set_grad_enabled(with_grad):
            feature_1, feature_2, feature_3 = self.model(imgs.to(self.device))
        embeddings = embeddings_concat(feature_1, feature_2)
        embeddings = embeddings_concat(embeddings, feature_3)
        embeddings = torch.index_select(
            embeddings,
            dim=1,
            index=self.embedding_ids,
        )
        return embeddings

    def _embed_batch_flatten(self, imgs, *args):
        embeddings = self._embed_batch(imgs, *args)
        _, C, _, _ = embeddings.shape
        return embeddings.permute(0, 2, 3, 1).reshape((-1, C))



import os
from os.path import join

import numpy as np
import torch

from src.anoseg.datasets.dataset import Dataset
from src.anoseg.models.PaDiM.backbone.padim import PaDiM


class Trainer(object):

    # region init

    def __init__(self, output_dir: str, dataset: Dataset, batch_size: int = 32, num_embeddings: int = 130,
                 backbone: str = "resnet18", image_size: int = 512, debugging: bool = False):
        self.device = 'cuda' if torch.cuda.is_available() else 'cpu'

        self.output_dir = output_dir
        self.dataset = dataset
        self.batch_size = batch_size
        self.image_size = image_size
        self.debugging = debugging

        self.padim = PaDiM(num_embeddings=num_embeddings, device=self.device, backbone=backbone,
                           size=(image_size, image_size))

    # endregion

    # region public methods

    def train(self) -> None:
        train_dataloader = self.dataset.get_train_dataloader(self.batch_size)

        self.padim.train(dataloader=train_dataloader, use_patches=False)
        N, means, covs, embedding_ids = self.padim.get_residuals()
        self.__save_model(N, means, covs, embedding_ids)

    # endregion

    # region private methods

    def __save_model(self, N, means, covs, embedding_ids) -> None:
        if not os.path.exists(self.output_dir):
            os.makedirs(self.output_dir)

        n_paths = join(self.output_dir, "n.npy")
        np.save(n_paths, N)

        means_paths = join(self.output_dir, "means.npy")
        np.save(means_paths, means)

        covs_paths = join(self.output_dir, "covs.npy")
        np.save(covs_paths, covs)

        embedding_ids_paths = join(self.output_dir, "embedding_ids.npy")
        np.save(embedding_ids_paths, embedding_ids)

    # endregion



import random

from src.anoseg.datasets.anomaly_creator.dfc_anomaly_creator import AnomalyCreator as DFCAnomalyCreator
from src.anoseg.datasets.anomaly_creator.ssaps_anomaly_creator import PatchAnomalyCreator
from src.anoseg.datasets.anomaly_creator.cutpaste_anomaly_creator import CutPaste

"""
    Options to create anomalies:
    
    DFC: Deep Feature Correspondence 
        paper: (https://www.sciencedirect.com/science/article/abs/pii/S0031320322003557)
    SSAPS: Self-Supervised Augmented Patches Segmentation for Anomaly Detection 
        paper: (https://openaccess.thecvf.com/content/ACCV2022/html/Long_Self-Supervised_Augmented_Patches_Segmentation_for_Anomaly_Detection_ACCV_2022_paper.html)
    CutPaste: Self-Supervised Learning for Anomaly Detection and Localization
        paper: https://arxiv.org/abs/2104.04015
"""


class AnomalyCreator(object):

    def __init__(self, img_size, mask_size, mean, std, imagenet_dir,
                 method='dfc', dfc_anomaly_size='big', cutpaste_mode='all'):
        self.img_size = img_size
        self.mask_size = mask_size
        self.mean = mean
        self.std = std
        self.anomaly_size = dfc_anomaly_size
        self.method = method

        if method == 'dfc':
            self.creator = DFCAnomalyCreator(img_size, mask_size, mean, std, imagenet_dir, dfc_anomaly_size)
        elif method == 'ssaps':
            self.creator = PatchAnomalyCreator()
        elif method == 'cutpaste':
            self.creator = CutPaste(mode=cutpaste_mode)
        elif method == 'all':
            self.cutpaste_creator = CutPaste(mode=cutpaste_mode)
            self.dfc_creator = DFCAnomalyCreator(img_size, mask_size, mean, std, imagenet_dir, dfc_anomaly_size)
            self.ssaps_creator = PatchAnomalyCreator()
        else:
            print("Unknown anomaly creation method.")

    def __call__(self, img):
        if self.method == 'all':
            creator = random.choice([self.cutpaste_creator, self.dfc_creator, self.ssaps_creator])
        else:
            creator = self.creator

        img_normal, img_abnormal, mask_normal, mask_abnormal = creator(img)

        return img_normal, img_abnormal, mask_normal, mask_abnormal



from typing import Tuple

import numpy as np
import torch
import os

from PIL import Image

from src.anoseg.datasets.dataset import Dataset
from src.anoseg.models.SPADE.base_spade.tester import Tester
from src.anoseg.models.SPADE.base_spade.trainer import Trainer
from src.anoseg.utils import visualization

"""
    Implementation of SPADE: Sub-Image Anomaly Detection with Deep Pyramid Correspondences
    Base implementation: https://github.com/byungjae89/SPADE-pytorch
    Paper: https://arxiv.org/abs/2005.02357
"""


class SPADE(object):

    # region init

    def __init__(self, model_path: str = None):
        self.device = 'cuda' if torch.cuda.is_available() else 'cpu'

        if model_path is None:
            self.trained = False
        else:
            self.model_path = model_path
            self.valid_model = self.__valid_model()
            self.trained = True

    # endregion

    # region public methods

    def train(self, dataset: Dataset, output_dir: str, debugging: bool = False, image_size: int = 256,
              batch_size: int = 32) -> None:
        trainer = Trainer(output_dir=output_dir,
                          dataset=dataset,
                          debugging=debugging,
                          batch_size=batch_size,
                          image_size=image_size)

        trainer.train()

    def eval(self, dataset: Dataset, debugging: bool = False, self_ensembling: bool = False,
             image_size: int = 256, mask_size: int = 1024,
             rot_90: bool = False, rot_180: bool = False, rot_270: bool = False,
             h_flip: bool = False, h_flip_rot_90: bool = False,
             h_flip_rot_180: bool = False, h_flip_rot_270: bool = False,
             integration_limit: float = 0.3, top_k: int = 5) -> None:
        if not self.trained:
            raise Exception("Model not trained.")
        if not self.valid_model:
            raise Exception("Invalid model.")

        tester = Tester(model_path=self.model_path, debugging=debugging,
                        image_size=image_size, mask_size=mask_size, use_self_ensembling=self_ensembling,
                        rot_90=rot_90, rot_180=rot_180, rot_270=rot_270, h_flip=h_flip,
                        h_flip_rot_90=h_flip_rot_90, h_flip_rot_180=h_flip_rot_180,
                        h_flip_rot_270=h_flip_rot_270,
                        integration_limit=integration_limit, top_k=top_k)

        tester.evaluate(dataset=dataset)

    def display_predictions(self, dataset: Dataset, debugging: bool = False, self_ensembling: bool = False,
                            image_size: int = 256, mask_size: int = 1024,
                            rot_90: bool = False, rot_180: bool = False, rot_270: bool = False,
                            h_flip: bool = False, h_flip_rot_90: bool = False,
                            h_flip_rot_180: bool = False, h_flip_rot_270: bool = False,
                            integration_limit: float = 0.3, top_k: int = 5) -> None:
        if not self.trained:
            raise Exception("Model not trained.")
        if not self.valid_model:
            raise Exception("Invalid model.")

        tester = Tester(model_path=self.model_path, debugging=debugging,
                        image_size=image_size, mask_size=mask_size, use_self_ensembling=self_ensembling,
                        rot_90=rot_90, rot_180=rot_180, rot_270=rot_270, h_flip=h_flip,
                        h_flip_rot_90=h_flip_rot_90, h_flip_rot_180=h_flip_rot_180,
                        h_flip_rot_270=h_flip_rot_270,
                        integration_limit=integration_limit, top_k=top_k)

        tester.display_predictions(dataset=dataset)

    def predict(self, image_path: str, display_prediction: bool, debugging: bool = False,
                mean: Tuple[float] = (0.485, 0.456, 0.406), std: Tuple[float] = (0.229, 0.224, 0.225),
                self_ensembling: bool = False, image_size: int = 256,
                rot_90: bool = False, rot_180: bool = False, rot_270: bool = False,
                h_flip: bool = False, h_flip_rot_90: bool = False,
                h_flip_rot_180: bool = False, h_flip_rot_270: bool = False,
                top_k: int = 5) -> Tuple[np.array, np.array]:
        if not self.trained:
            raise Exception("Model not trained.")
        if not self.valid_model:
            raise Exception("Invalid model.")

        tester = Tester(model_path=self.model_path, debugging=debugging,
                        image_size=image_size, use_self_ensembling=self_ensembling,
                        rot_90=rot_90, rot_180=rot_180, rot_270=rot_270, h_flip=h_flip,
                        h_flip_rot_90=h_flip_rot_90, h_flip_rot_180=h_flip_rot_180,
                        h_flip_rot_270=h_flip_rot_270, top_k=top_k)

        score, binary_score = tester.predict(image_path=image_path, mean=mean, std=std)

        if display_prediction:
            original = Image.open(image_path).convert('RGB')

            visualization.display_images(img_list=[original, score, binary_score],
                                         titles=['original', 'score', 'binary_score'],
                                         cols=3)

        return score, binary_score

    # endregion

    # region private methods

    def __valid_model(self) -> bool:
        valid = True
        valid = valid and os.path.exists(os.path.join(self.model_path, "avgpool.npy"))
        valid = valid and os.path.exists(os.path.join(self.model_path, "layer_1.npy"))
        valid = valid and os.path.exists(os.path.join(self.model_path, "layer_2.npy"))
        valid = valid and os.path.exists(os.path.join(self.model_path, "layer_3.npy"))

        return valid

    # endregion



from typing import Tuple, Union, List
from functools import reduce
from itertools import product

import cv2
import numpy as np
from numpy import ndarray as NDArray
import torch

# A box defined as (x1, y1, x2, y2)
Region = Tuple[int, int, int, int]


def propose_regions_cv2(patches, threshold: float = 0.75, **kwargs):
    """
    Faster region proposal

    >>> propose_regions_cv2(np.array([[1, 0, 1], [1, 0, 1], [0, 1, 0]]))
    [(0, 0, 1, 2, 1), (2, 0, 3, 2, 1), (1, 2, 2, 3, 1)]

    >>> propose_regions_cv2(np.array([[0, 0], [0, 0]]))
    []

    """
    if isinstance(patches, torch.Tensor):
        patches = patches.cpu().numpy()
    mask = (patches >= threshold).astype(np.int8)

    _, _, stats, _ = cv2.connectedComponentsWithStats(mask, connectivity=4)

    def stat_to_box(stat):
        x1, y1, w, h, _ = stat
        return (x1, y1, x1 + w, y1 + h, 1)
    boxes = map(stat_to_box, stats[1:])
    return filter_regions(boxes, **kwargs)


def propose_region(
    patches: NDArray,
    threshold: float = 0.75,
    explore_diagonals: bool = False,
) -> Union[None, Region]:
    """Proposes regions from a set of patch activations

    >>> propose_region(np.array([[1, 1, 0], [0, 1, 0], [0, 0, 0]]))
    (0, 0, 2, 2, 1)

    >>> propose_region(np.array([[1, 0], [0, 0]]))
    (0, 0, 1, 1, 1)

    >>> propose_region(np.array([[1, 0], [0, 1]]), explore_diagonals=True)
    (0, 0, 2, 2, 1)

    >>> propose_region(np.array([[.1, 0], [0, 0]]), threshold=0.2)

    Params
    ======
        patches: ndarray - (h * w) patch activations
        threshold: float - the activation threshold
    Returns
    =======
        box - the bounding box
    """
    h, w = patches.shape

    first_patch_idx = np.argmax(patches, axis=None)
    px, py = first_patch_idx % w, first_patch_idx // w
    patch_value = patches[py, px]

    if patch_value < threshold:
        return None
    patches[py, px] = 0.0

    directions = [
        (0, 1),  # right
        (0, -1),  # left
        (1, 0),  # down
        (-1, 0),  # up
    ]
    if explore_diagonals:
        directions.extend([
            (-1, -1),  # upper left
            (-1, 1),  # bottom left
            (1, -1),  # upper right
            (1, 1),  # bottom right
        ])

    def explore_neighbors(x, y):
        neighbors = []
        for dx, dy in directions:
            nx, ny = dx + x, dy + y
            # invalid directions
            if nx not in range(w) or ny not in range(h):
                continue
            # patch is activated
            if patches[ny, nx] >= threshold:
                neighbors.append((nx, ny))
                patches[ny, nx] = 0.0

        new_neighbors = []
        for neighbor in neighbors:
            new_neighbors.extend(explore_neighbors(*neighbor))

        neighbors.extend(new_neighbors)
        return neighbors

    neighbors = explore_neighbors(px, py)
    neighbors.append((px, py))
    min_x = reduce(min, map(lambda p: p[0], neighbors), w)
    min_y = reduce(min, map(lambda p: p[1], neighbors), h)
    max_x = reduce(max, map(lambda p: p[0], neighbors), 0) + 1
    max_y = reduce(max, map(lambda p: p[1], neighbors), 0) + 1

    return (min_x, min_y, max_x, max_y, patch_value)


def propose_regions(patches: NDArray,
                    threshold: float = 0.75,
                    **kwargs) -> List[Region]:
    """Proposes many regions

    >>> propose_regions(np.array([[1, 0, 1], [1, 0, 1], [0, 1, 0]]))
    [(0, 0, 1, 2, 1), (2, 0, 3, 2, 1), (1, 2, 2, 3, 1)]

    >>> propose_regions(np.array([[0, 0], [0, 0]]))
    []

    Params
    ======
        patches: ndarray - (h * w) patch activations
        threshold: float - the activation threshold
    Returns
    =======
        boxes: List[Region] - (x1, y1, x2, y2) bounding boxes
    """
    regions = []
    proposal = propose_region(patches, threshold)
    while proposal is not None:
        regions.append(proposal)
        proposal = propose_region(patches, threshold)

    return filter_regions(regions, **kwargs)


def filter_regions(
    regions: List[Region],
    min_area: int = 1,
    use_nms: bool = False,
    **kwargs,
) -> List[Region]:
    """Filters out regions that are not relevant

    >>> filter_regions([(0, 0, 1, 1)], min_area = 2)
    []

    >>> filter_regions([(0, 0, 1, 1), (0, 0, 2, 1)], min_area = 2)
    [(0, 0, 2, 1)]

    Params
    ======
        regions: List[Tuple[float, float, float, float]] - regions
        min_area: int - the minimum area for a region
    Returns
    =======
        regions: List[Tuple[float, float, float, float]] - regions
    """
    def get_area(region) -> float:
        x1, y1, x2, y2 = region[:4]
        return abs(x1 - x2) * abs(y1 - y2)

    if use_nms:
        regions = non_maximum_suppression(regions, **kwargs)
    return list(filter(lambda r: get_area(r) >= min_area, regions))


def _serialize_region(region: Region) -> List[Tuple[int, int]]:
    """Lists the patches that are part of a region

    >>> _serialize_region((0, 0, 1, 2))
    [(0, 0), (0, 1)]

    Params
    ======
        region - Region the region to serialize
    Returns
    =======
        patches - List[Tuple[int, int]] patches part of the region
    """
    x1, y1, x2, y2 = region
    return list(product(range(x1, x2), range(y1, y2)))


def IoU(r1: Region, r2: Region) -> float:
    """Intersection over Union

    >>> IoU((0, 0, 1, 2), (0, 0, 1, 1))
    0.5

    Params
    ======
        r1 - Region
        r2 - Region
    Returns
    =======
        IoU - the insersection over union
    """
    s1 = _serialize_region(r1)
    s2 = _serialize_region(r2)

    intersection = set(s1).intersection(s2)
    union = set(s1).union(s2)

    return len(intersection) / len(union)


def floating_IoU(r1: Tuple[float, float, float, float],
                 r2: Tuple[float, float, float, float]) -> float:
    """IoU for non-grid based boxes

    >>> floating_IoU((0, 0, 1, 1), (0, 0, 1, 2))
    0.5

    Params
    ======
        r1 - (x, y, w, h,...)
        r2 - (x, y, w, h,...)
    Returns
    =======
        IoU - the insersection over union
    """
    x1, y1, w1, h1 = r1[:4]
    x2, y2, w2, h2 = r2[:4]

    left = max(x1, x2)
    right = min(x1 + w1, x2 + w2)
    top = max(y1, y2)
    bottom = min(y1 + h1, y2 + h2)

    overlap_height = bottom - top
    overlap_width = right - left

    if overlap_height <= 0 or overlap_width <= 0:
        overlap_area = 0
    else:
        overlap_area = overlap_height * overlap_width

    area1 = w1 * h1
    area2 = w2 * h2

    union_area = area1 + area2 - overlap_area
    return overlap_area / union_area


def cvt_xyxys_xywhs(box):
    x1, y1, x2, y2, s = box
    return (x1, y1, x2 - x1, y2 - y1, s)


def non_maximum_suppression(boxes: List, iou_threshold: float = 0.5) -> List:
    """Filter boxes using non-maximum suppression

    >>> non_maximum_suppression([(0, 0, 1, 2, .2), (0, 0, 1, 1, .4)])
    [(0, 0, 1, 1, 0.4)]

    Params
    ======
        boxes: List[Regions] - regions with score
        iou_threshold: float - the threshold at which to remove boxes
    Returns
    =======
        boxes - List[Regions] - the filtered regions
    """
    # TODO: Fast version using numpy
    # Sort boxes by confidence scores
    boxes = sorted(boxes, key=lambda b: b[4])
    new_boxes = []
    while len(boxes) > 0:
        box = boxes.pop()
        new_boxes.append(box)
        # remove boxes
        boxes = list(
            filter(
                lambda other_box: floating_IoU(cvt_xyxys_xywhs(
                    box), cvt_xyxys_xywhs(other_box)) < iou_threshold, boxes))

    return new_boxes


if __name__ == "__main__":
    import doctest
    doctest.testmod()

    import timeit

    array = 0.5 + np.random.randn(104, 104)

    def test_cv2():
        n_array = np.copy(array)
        propose_regions_cv2(n_array, treshold=0.5, min_area=1)

    def test_orig():
        n_array = np.copy(array)
        propose_regions(n_array, treshold=0.5, min_area=1)

    n_try = 100
    print("starting cv2")
    time_cv2 = timeit.timeit(test_cv2, number=n_try)
    print("cv2 done, starting orig")
    time_orig = timeit.timeit(test_orig, number=n_try)
    print("orig done")

    print(f"time_cv2 = {time_cv2:.5f}")
    print(f"time_cv2 (unit) = {time_cv2 / n_try:.5f}")
    print(f"time_orig = {time_orig:.5f}")
    print(f"time_orig (unit) = {time_orig / n_try:.5f}")



import cv2
import numpy as np
from skimage.transform import resize

from src.anoseg.datasets.imagenet import ILSVRC

"""
    Implementation based on https://github.com/YoungGod/DFC/tree/main
    Paper: https://www.sciencedirect.com/science/article/abs/pii/S0031320322003557
"""


class AnomalyCreator(object):

    # region init

    def __init__(self, img_size, mask_size, mean, std, imagenet_dir, anomaly_size='all'):
        self.img_size = img_size
        self.mask_size = mask_size
        self.mean = mean
        self.std = std
        self.anomaly_size = anomaly_size

        self.__load_imagenet(imagenet_dir)
        if self.anomaly_size != "all":
            self.__set_mask_variables(anomaly_size)

    def __load_imagenet(self, imagenet_dir):
        if imagenet_dir is None:
            print("Imagenet dir not specified! Traning witout Self-Feature Enhancement with external datasets")
            self.aug_dataset = None
        else:
            try:
                self.aug_dataset = ILSVRC(ilsvrc_data_path=imagenet_dir, mean=self.mean, std=self.std, val=False)
            except Exception as ex:
                print("Imagenet was not Found! Traning witout Self-Feature Enhancement with external datasets")
                self.aug_dataset = None

    # endregion

    # region anomaly creation

    """
        src: https://github.com/YoungGod/DFC
        code was slightly adjusted

        OE_mode 0: set to black
        OE_mode 1: inpainting with opencv
        OE_mode 2: filling with sampled pixel from the original image
        OE_mode 3: filling with noise
        OE_mode 4: filling with imagenet
    """

    def __call__(self, input_img):
        if self.anomaly_size == "all":
            self.__set_mask_variables(self.anomaly_size)

        input_img = np.asarray(input_img)
        input_img = cv2.resize(input_img, dsize=(self.img_size, self.img_size), interpolation=cv2.INTER_LINEAR)
        img_normal = input_img.copy()
        img_abnormal = input_img.copy()

        # OE
        if self.aug_dataset is not None:
            oe_mode = np.random.randint(0, 5)  # OE mode
        else:
            oe_mode = np.random.randint(0, 4)  # OE mode

        # generate random anomaly mask
        original_mask = self.__generate_mask()  # [0, 1]
        rescaled_mask = resize(original_mask, (self.img_size, self.img_size), anti_aliasing=False)

        if oe_mode == 0:
            img_abnormal[rescaled_mask == 1] = 0

        # inpainting with opencv
        if oe_mode == 1:
            inpaint_mask = (rescaled_mask * 255).astype(np.uint8)
            img_abnormal = cv2.inpaint(img_abnormal, inpaint_mask, 5, cv2.INPAINT_TELEA)

        # constant fill (filling with sampled pixel from the orginal image)
        if oe_mode == 2:
            indx, indy = np.where(
                rescaled_mask > 0 if np.random.rand() > 0.4 else rescaled_mask == 0)  # prefer to sample pixel in anoamly region
            index = np.random.randint(0, len(indx))
            img_to_inpaint = np.ones_like(img_abnormal) * img_abnormal[indx[index], indy[index]]
            # alpha = 0.1 * np.random.randint(1,5)    # blending coefficient
            alpha_min = 0.1
            alpha_range = 0.4
            alpha = alpha_min + np.random.rand() * alpha_range
            img_to_inpaint = cv2.addWeighted(img_abnormal, alpha, img_to_inpaint, 1 - alpha, 0)
            img_abnormal = img_to_inpaint * rescaled_mask[:, :, np.newaxis] + img_abnormal * (
                    1 - rescaled_mask[:, :, np.newaxis])

        # noise fill
        if oe_mode == 3:
            img_to_inpaint = (np.random.rand(*img_abnormal.shape) * 255).astype(np.int16)
            img_to_inpaint = img_to_inpaint.astype(np.uint8)
            alpha_min = 0.3
            alpha_range = 0.5
            alpha = alpha_min + np.random.rand() * alpha_range
            img_to_inpaint = cv2.addWeighted(img_abnormal, alpha, img_to_inpaint, 1 - alpha, 0)
            img_abnormal = img_to_inpaint * rescaled_mask[:, :, np.newaxis] + img_abnormal * (
                    1 - rescaled_mask[:, :, np.newaxis])

        # a-blending
        if oe_mode == 4:
            # generate a image for blending (here from imagenet)
            img_to_blend = self.__generate_image_to_blend()
            alpha_min = 0.3
            alpha_range = 0.5
            alpha = alpha_min + np.random.rand() * alpha_range
            img_to_blend = cv2.addWeighted(img_abnormal, alpha, img_to_blend, 1 - alpha, 0)
            img_abnormal = img_to_blend * rescaled_mask[:, :, np.newaxis] + img_abnormal * (
                    1 - rescaled_mask[:, :, np.newaxis])

        mask_normal = np.zeros((self.mask_size, self.mask_size), dtype=np.float32)
        mask_abnormal = original_mask.astype(np.float32)

        # visualization.display_images([original_mask, rescaled_mask], ['original', 'rescaled'])
        img_normal = img_normal.astype(np.uint8)
        img_abnormal = img_abnormal.astype(np.uint8)

        return img_normal, img_abnormal, mask_normal, mask_abnormal

    def __set_mask_variables(self, size):
        if size == 'big':
            self.max_brush_width_lowervalue = self.mask_size // 32
            self.max_vertex_lowervalue = self.mask_size // 21
            self.max_brush_width_uppervalue = self.mask_size // 5
            self.max_vertex_uppervalue = self.mask_size // 16
            self.max_length = self.mask_size // 8
        elif size == 'medium':
            self.max_brush_width_lowervalue = self.mask_size // 64
            self.max_vertex_lowervalue = self.mask_size // 42
            self.max_brush_width_uppervalue = self.mask_size // 10
            self.max_vertex_uppervalue = self.mask_size // 32
            self.max_length = self.mask_size // 16
        else:
            self.max_brush_width_lowervalue = self.mask_size // 128
            self.max_vertex_lowervalue = self.mask_size // 84
            self.max_brush_width_uppervalue = self.mask_size // 20
            self.max_vertex_uppervalue = self.mask_size // 64
            self.max_length = self.mask_size // 32

    def __generate_mask(self):
        # max_brush_width_uppervalue = 48
        # max_vertex_uppervalue = 16

        max_anomaly_regions = 3  # possible max number of anomaly regions
        parts = np.random.randint(1, max_anomaly_regions + 1)

        mask = 0
        for part in range(parts):
            # brush params
            max_brush_width = np.random.randint(self.max_brush_width_lowervalue, self.max_brush_width_uppervalue)
            # max_length = 32
            max_vertex = np.random.randint(self.max_vertex_lowervalue, self.max_vertex_uppervalue)

            temp_mask = self.__np_free_form_mask(max_vertex, self.max_length, max_brush_width, max_angle=180,
                                                 h=self.mask_size, w=self.mask_size)
            mask = np.logical_or(mask, temp_mask)
        return mask

    def __generate_image_to_blend(self):
        idx = int(np.random.rand() * len(self.aug_dataset))
        img_path = self.aug_dataset.img_paths[idx]
        img_to_blend = cv2.imread(img_path)
        img_to_blend = cv2.resize(img_to_blend, (self.img_size, self.img_size))
        img_to_blend = cv2.cvtColor(img_to_blend, cv2.COLOR_BGR2RGB)

        return img_to_blend

    def __np_free_form_mask(self, max_vertex, max_length, max_brush_width, max_angle, h, w):
        # start_value = self.mask_size // 13
        start_value = 10
        next_value = self.mask_size // 6

        mask = np.zeros((h, w), np.float32)
        num_vertex = np.random.randint(2, max_vertex + 1)
        start_y = np.random.randint(start_value, h - start_value)
        start_x = np.random.randint(start_value, w - start_value)
        brush_width = 0
        pre_angle = 0
        for i in range(num_vertex):
            angle = np.random.randint(max_angle + 1)
            angle = (angle / 360.0 * 2 * np.pi + pre_angle) / 2

            if i % 2 == 0:
                angle = 2 * np.pi - angle
            length = np.random.randint(1, max_length + 1)
            brush_width = np.random.randint(self.max_brush_width_lowervalue, max_brush_width + 1) // 2 * 2
            next_y = start_y + length * np.cos(angle)
            next_x = start_x + length * np.sin(angle)

            next_y = np.maximum(np.minimum(next_y, h - next_value), next_value).astype(int)
            next_x = np.maximum(np.minimum(next_x, w - next_value), next_value).astype(int)

            cv2.line(mask, (start_y, start_x), (next_y, next_x), 1, brush_width)

            start_y, start_x = next_y, next_x
            pre_angle = angle

        return mask

    # endregion



import random
from typing import List, Tuple
from PIL import Image
from torch.utils.data import Dataset
from torchvision.transforms import transforms

from src.anoseg.datasets.anomaly_creator.anomaly_creator import AnomalyCreator

"""
    Dataset to train DFC SPADE or PaDiM model
    Data can be augmented
"""


class TrainDataset(Dataset):

    def __init__(self, image_paths: List[str], imagenet_dir: str, img_size: int = 256, mask_size: int = 1024,
                 rot_90: bool = False, rot_180: bool = False, rot_270: bool = False, h_flip: bool = False,
                 h_flip_rot_90: bool = False, h_flip_rot_180: bool = False, h_flip_rot_270: bool = False,
                 self_supervised_training: bool = True,
                 mean: Tuple[float] = (0.485, 0.456, 0.406), std: Tuple[float] = (0.229, 0.224, 0.225),
                 dfc_anomaly_size: str = 'big', method: str = 'dfc', cutpaste_mode: str = 'all'):

        self.image_paths = image_paths
        self.images = []
        self.normal_images = []
        self.img_size = img_size
        self.mask_size = mask_size
        self.len = len(self.image_paths)
        self.mean = list(mean)
        self.std = list(std)

        self.rot_90 = rot_90
        self.rot_180 = rot_180
        self.rot_270 = rot_270
        self.h_flip = h_flip
        self.h_flip_rot_90 = h_flip_rot_90
        self.h_flip_rot_180 = h_flip_rot_180
        self.h_flip_rot_270 = h_flip_rot_270

        self.self_supervised_training = self_supervised_training

        for image_path in image_paths:
            img = Image.open(image_path).convert('RGB')
            img = img.resize((self.img_size, self.img_size))
            self.images.append(img)

        self.__build_transforms()
        self.__build_aug_transforms()

        if self.self_supervised_training:
            self.anomaly_creator = AnomalyCreator(img_size=img_size, mask_size=mask_size, mean=self.mean, std=self.std,
                                                  imagenet_dir=imagenet_dir, method=method,
                                                  dfc_anomaly_size=dfc_anomaly_size, cutpaste_mode=cutpaste_mode)

    def __getitem__(self, index):
        img = self.images[index]
        augmented_img = self.__augment_img(img)

        if self.self_supervised_training:
            img_normal, img_abnormal, mask_normal, mask_abnormal = \
                self.anomaly_creator(augmented_img)

            img_abnormal = Image.fromarray(img_abnormal)

            img_normal = self.transform(img_normal)
            img_abnormal = self.transform(img_abnormal)

            return img_normal, img_abnormal, mask_normal, mask_abnormal
        else:
            return self.transform(augmented_img)

    def __len__(self):
        return self.len

    # region private methods

    def __build_transforms(self):
        normalize = transforms.Normalize(mean=self.mean, std=self.std)
        self.transform = transforms.Compose([transforms.ToTensor(), normalize])

    def __build_aug_transforms(self) -> None:
        self.possible_transforms: List[transforms.Compose] = [transforms.Compose([])]

        if self.rot_90:
            trans = transforms.Compose([transforms.RandomRotation(degrees=[90, 90])])
            self.possible_transforms.append(trans)
        if self.rot_180:
            trans = transforms.Compose([transforms.RandomRotation(degrees=[180, 180])])
            self.possible_transforms.append(trans)
        if self.rot_270:
            trans = transforms.Compose([transforms.RandomRotation(degrees=[270, 270])])
            self.possible_transforms.append(trans)
        if self.h_flip:
            trans = transforms.Compose([transforms.RandomHorizontalFlip(p=1)])
            self.possible_transforms.append(trans)
        if self.h_flip_rot_90:
            trans = transforms.Compose([transforms.RandomHorizontalFlip(p=1),
                                        transforms.RandomRotation(degrees=[90, 90])])
            self.possible_transforms.append(trans)
        if self.h_flip_rot_180:
            trans = transforms.Compose([transforms.RandomHorizontalFlip(p=1),
                                        transforms.RandomRotation(degrees=[180, 180])])
            self.possible_transforms.append(trans)
        if self.h_flip_rot_270:
            trans = transforms.Compose([transforms.RandomHorizontalFlip(p=1),
                                        transforms.RandomRotation(degrees=[270, 270])])
            self.possible_transforms.append(trans)

    def __augment_img(self, img: Image) -> Image:
        aug_value: int = int(random.uniform(0, len(self.possible_transforms)))
        augmented = self.possible_transforms[aug_value](img)

        return augmented

    # endregion



import math
import torchvision
import torchvision.transforms.functional as TF
import numpy as np
import torch
import torch.nn.functional as F
from abc import ABC
from typing import List
from os.path import join
from PIL import Image
from torch import Tensor
from torchvision.transforms import transforms

from src.anoseg.models.PaDiM.backbone.padim import PaDiM
from src.anoseg.models.utils import BaseTester


"""
    Implementation of PaDiM: a Patch Distribution Modeling Framework for Anomaly Detection and Localization
    Code modified from https://github.com/Pangoraw/PaDiM
    Paper: https://arxiv.org/abs/2011.08785
"""


class Tester(BaseTester, ABC):

    # region init

    def __init__(self, model_path: str, debugging: bool = False,
                 image_size: int = 256, mask_size: int = 1024, use_self_ensembling: bool = False,
                 rot_90: bool = False, rot_180: bool = False, rot_270: bool = False, h_flip: bool = False,
                 h_flip_rot_90: bool = False, h_flip_rot_180: bool = False, h_flip_rot_270: bool = False,
                 integration_limit: float = 0.3, backbone: str = "wide_resnet50"):
        self.device = 'cuda' if torch.cuda.is_available() else 'cpu'

        super().__init__(model_path=model_path, debugging=debugging, image_size=image_size, mask_size=mask_size,
                         use_self_ensembling=use_self_ensembling, rot_90=rot_90, rot_180=rot_180, rot_270=rot_270,
                         h_flip=h_flip, h_flip_rot_90=h_flip_rot_90, h_flip_rot_180=h_flip_rot_180,
                         h_flip_rot_270=h_flip_rot_270, integration_limit=integration_limit)
        self.backbone = backbone

        self.__load_model()

    def __load_model(self) -> None:
        n = np.load(join(self.model_path, "n.npy")).item()
        means = np.load(join(self.model_path, "means.npy"))
        covs = np.load(join(self.model_path, "covs.npy"))
        embedding_ids = np.load(join(self.model_path, "embedding_ids.npy"))
        padim = PaDiM.from_residuals(N=n,
                                     means=means,
                                     covs=covs,
                                     embedding_ids=embedding_ids,
                                     backbone=self.backbone,
                                     device=self.device,
                                     img_size=self.image_size)
        self.padim = padim

    # end region

    # region implement abstract methods

    def score(self, img_input) -> np.array:  # returns score with shape (1024, 1024)
        distances = self.padim.predict(img_input)
        w = int(math.sqrt(distances.numel()))
        raw_score = distances.reshape(1, 1, w, w)
        raw_score = F.interpolate(raw_score, size=(self.mask_size, self.mask_size), mode="bilinear",
                                  align_corners=True)

        raw_score = raw_score.detach().cpu().numpy().squeeze()

        return raw_score

    def score_with_augmentation(self, img_input) -> np.array:
        score_list = self._get_self_ensembling_scores(img_input)
        final_score = self._combine_scores(score_list)

        return final_score

    def preprocess_img(self, image_path: str, mean: List[float], std: List[float]) -> Tensor:
        original = Image.open(image_path).convert('RGB')

        normalize = transforms.Normalize(mean=mean, std=std)
        resize = torchvision.transforms.Resize(size=self.image_size, interpolation=TF.InterpolationMode.BILINEAR,
                                               antialias=True)

        transform = transforms.Compose([transforms.ToTensor(), normalize, resize])
        preprocessed = transform(original)

        preprocessed = preprocessed[None, :]

        return preprocessed

    # endregion



import torch
import os
import pandas as pd
import torch.nn as nn
from torch.optim import Adam
from contextlib import contextmanager
from torchvision.transforms import Resize, InterpolationMode
from typing_extensions import Literal

from src.anoseg.datasets.dataset import Dataset
from src.anoseg.models.DFC.backbone.vgg19 import VGG19
from src.anoseg.models.DFC.backbone.vgg19_s import VGG19_S


@contextmanager
def task(_):
    yield


class Trainer(object):

    # region init

    _DATASET_TYPES = Literal["textures", "objects"]
    _CNN_LAYERS_TEXTURES = ("relu4_1", "relu4_2", "relu4_3", "relu4_4")
    _CNN_LAYERS_OBJECTS = ("relu4_3", "relu4_4", "relu5_1", "relu5_2")

    def __init__(self, output_dir: str, dataset: Dataset, dataset_type: _DATASET_TYPES, batch_size: int = 8,
                 n_epochs: int = 201, lr: float = 2e-4, train_split: float = 0.95,
                 pretrained_weights_dir: str = None, imagenet_dir: str = None, early_stopping: bool = True,
                 patience: int = 10, debugging: bool = True):
        self.device = 'cuda' if torch.cuda.is_available() else 'cpu'

        self.output_dir = output_dir
        self.dataset = dataset
        self.dataset_type = dataset_type
        self.batch_size = batch_size
        self.epochs = n_epochs
        self.learning_rate = lr
        self.train_split = train_split
        self.pretrained_weights_dir = pretrained_weights_dir
        self.imagenet_dir = imagenet_dir
        self.early_stopping = early_stopping
        self.patience = patience
        self.debugging = debugging

        self.epochs_without_improvement = 0
        self.best_val_loss = float("inf")

        self.__build_models()
        self.loss = nn.MSELoss(reduction='mean')

        self.loss_df = pd.DataFrame({'epoch': [], 'loss': [], 'loss_normal': [], 'loss_abnormal': []})
        self.val_loss_df = pd.DataFrame({'epoch': [], 'loss': [], 'loss_normal': [], 'loss_abnormal': []})
        self.eval_df = pd.DataFrame({'epoch': [], 'roc': [], 'pro': [], 'iou': []})

        self.__log_message("Training Patch DFC model")
        self.__log_message(f"Using device: {self.device}")
        self.__log_message(f"Saving model in {self.output_dir}")

        self.__init_feat_layers()

        self.__log_message(f"Batch size: {self.batch_size}")
        self.__log_message(f"Epochs: {self.epochs}")

    def __build_models(self):
        self.feature_extraction = self.__build_feature_extractor()

        self.feature_matching_big = self.__build_feature_matching()
        self.feature_matching_medium = self.__build_feature_matching()
        self.feature_matching_small = self.__build_feature_matching()

        self.optimizer_big = Adam(self.feature_matching_big.parameters(), lr=self.learning_rate, weight_decay=1e-5)
        self.optimizer_medium = Adam(self.feature_matching_medium.parameters(), lr=self.learning_rate,
                                     weight_decay=1e-5)
        self.optimizer_small = Adam(self.feature_matching_small.parameters(), lr=self.learning_rate, weight_decay=1e-5)

    def __build_feature_extractor(self):
        res = VGG19(pretrain=True, gradient=False, pool='avg',
                    pretrained_weights_dir=self.pretrained_weights_dir).to(self.device)
        return res

    def __build_feature_matching(self):
        res = VGG19_S(pretrain=False, gradient=True, pool='avg').to(self.device)
        return res

    def __init_feat_layers(self):
        if self.dataset_type == "objects":
            self.feat_layers = self._CNN_LAYERS_OBJECTS
            self.__log_message("Using object layers.")
        elif self.dataset_type == "textures":
            self.feat_layers = self._CNN_LAYERS_TEXTURES
            self.__log_message("Using texture layers.")
        else:
            raise Exception("Unknown dataset type. Valid types: ['textures', 'objects']")

    # endregion

    # region public methods

    def train(self):
        self.feature_extraction.eval()

        if os.path.exists(os.path.join(self.output_dir, 'big', 'final')):
            self.__log_message("Big model already exists. Skipped training big model.")
        else:
            self.__train_big_patches()
        if os.path.exists(os.path.join(self.output_dir, 'medium', 'final')):
            self.__log_message("Medium model already exists. Skipped training medium model.")
        else:
            self.__train_medium_patches()
        if os.path.exists(os.path.join(self.output_dir, 'small', 'final')):
            self.__log_message("Small model already exists. Skipped training small model.")
        else:
            self.__train_small_patches()

    # endregion

    # region training

    def __train_big_patches(self):
        with task("dataset"):
            train_loader, val_loader = self.dataset.get_train_and_val_dataloader(self.batch_size, self.train_split)
            self.__log_message(f"Using {len(train_loader.dataset)} images for training")

        with task("train"):
            self.feature_extraction.eval()
            self.feature_matching_big.train()

            self.__log_message("Start training with big patches...")
            loss_avg = 0.
            loss_normal_avg = 0.
            loss_abnormal_avg = 0.

            loss_save_path = os.path.join(self.output_dir, "big", "loss.csv")
            val_loss_save_path = os.path.join(self.output_dir, "big", "val_loss.csv")
            eval_save_path = os.path.join(self.output_dir, "eval.csv")

            best_loss = float("inf")

            for epoch in range(self.epochs):
                self.feature_matching_big.train()

                for normal, abnormal, normal_mask, abnormal_mask in train_loader:
                    normal = normal.to(self.device)
                    abnormal = abnormal.to(self.device)

                    self.optimizer_big.zero_grad()

                    with task('normal'):
                        surrogate_label_normal = self.feature_extraction(normal, self.feat_layers)
                        pred = self.feature_matching_big(normal, self.feat_layers)
                        loss_normal = 0
                        for feat_layer in self.feat_layers:
                            loss_normal += self.loss(pred[feat_layer], surrogate_label_normal[feat_layer])
                        loss_normal = loss_normal / len(self.feat_layers)

                    with task('abnormal'):
                        pred = self.feature_matching_big(abnormal, self.feat_layers)
                        loss_abnormal = 0
                        for feat_layer in self.feat_layers:
                            loss_abnormal += self.loss(pred[feat_layer], surrogate_label_normal[feat_layer])
                        loss_abnormal = loss_abnormal / len(self.feat_layers)

                    alpha = 1
                    loss = loss_normal + alpha * loss_abnormal
                    loss.backward()
                    self.optimizer_big.step()

                    # exponential moving average
                    loss_avg = loss_avg * 0.9 + float(loss.item()) * 0.1
                    loss_normal_avg = loss_normal_avg * 0.9 + float(loss_normal.item()) * 0.1
                    loss_abnormal = alpha * loss_abnormal
                    loss_abnormal_avg = loss_abnormal_avg * 0.9 + float(loss_abnormal.item()) * 0.1

                if loss_avg < best_loss:
                    best_loss = loss_avg
                    self.__save_model("big/best", self.feature_matching_big)
                if epoch % 1 == 0:
                    self.__log_message(f"Epoch {epoch}, loss = {loss_avg:.5f}, loss_normal = {loss_normal_avg:.5f}, "
                                       f"loss_abnormal = {loss_abnormal_avg:.5f}")
                    self.loss_df.loc[len(self.loss_df)] = [epoch, loss_avg, loss_normal_avg, loss_abnormal_avg]
                    val_loss = self.__calc_validation_loss(val_loader, epoch, self.feature_matching_big)
                if epoch % 10 == 0 and epoch != 0:
                    self.__save_model("big/epoch_{}".format(epoch), self.feature_matching_big)
                    self.loss_df.to_csv(loss_save_path, index=False)
                    self.val_loss_df.to_csv(val_loss_save_path, index=False)
                if epoch > 100:
                    if self.early_stopping and self.__stop_early(val_loss):
                        break
                if epoch == 100:
                    self.__set_new_lr(self.optimizer_big)

                torch.cuda.empty_cache()

        # save model
        self.__save_model("big/final", self.feature_matching_big)

        self.loss_df.to_csv(loss_save_path, index=False)
        self.val_loss_df.to_csv(val_loss_save_path, index=False)
        self.eval_df.to_csv(eval_save_path, index=False)

        self.__log_message("Matching Net for big patches Trained.")

        self.best_val_loss = float("inf")
        self.epochs_without_improvement = 0

    def __train_medium_patches(self):
        with task("dataset"):
            train_loader, val_loader = self.dataset.get_medium_patches_train_and_val_dataloader(self.batch_size,
                                                                                                self.train_split)

        with task("train"):
            self.feature_extraction.eval()
            self.feature_matching_medium.train()

            self.__log_message("Started training medium patches...")
            loss_avg = 0.
            loss_normal_avg = 0.
            loss_abnormal_avg = 0.

            loss_save_path = os.path.join(self.output_dir, "medium", "loss.csv")
            val_loss_save_path = os.path.join(self.output_dir, "medium", "val_loss.csv")

            best_loss = float("inf")

            for epoch in range(self.epochs):
                self.feature_matching_medium.train()

                for normals, abnormals, normal_masks, abnormal_masks in train_loader:  # for all train images
                    for normal, abnormal, normal_mask, abnormal_mask in zip(normals, abnormals, normal_masks,
                                                                            abnormal_masks):
                        normal = normal.to(self.device)
                        abnormal = abnormal.to(self.device)

                        self.optimizer_medium.zero_grad()

                        with task('normal'):
                            surrogate_label_normal = self.feature_extraction(normal, self.feat_layers)
                            pred = self.feature_matching_medium(normal, self.feat_layers)
                            loss_normal = 0
                            for feat_layer in self.feat_layers:
                                loss_normal += self.loss(pred[feat_layer], surrogate_label_normal[feat_layer])
                            loss_normal = loss_normal / len(self.feat_layers)

                        with task('abnormal'):
                            pred = self.feature_matching_medium(abnormal, self.feat_layers)
                            loss_abnormal = 0
                            for feat_layer in self.feat_layers:
                                loss_abnormal += self.loss(pred[feat_layer], surrogate_label_normal[feat_layer])
                            loss_abnormal = loss_abnormal / len(self.feat_layers)

                        alpha = 1
                        loss = loss_normal + alpha * loss_abnormal
                        loss.backward()
                        self.optimizer_medium.step()

                        # exponential moving average
                        loss_avg = loss_avg * 0.9 + float(loss.item()) * 0.1
                        loss_normal_avg = loss_normal_avg * 0.9 + float(loss_normal.item()) * 0.1
                        loss_abnormal = alpha * loss_abnormal
                        loss_abnormal_avg = loss_abnormal_avg * 0.9 + float(loss_abnormal.item()) * 0.1

                if loss_avg < best_loss:
                    best_loss = loss_avg
                    self.__save_model("medium/best", self.feature_matching_medium)
                if epoch % 1 == 0:
                    self.__log_message(f"Epoch {epoch}, loss = {loss_avg:.5f}, loss_normal = {loss_normal_avg:.5f}, "
                                       f"loss_abnormal = {loss_abnormal_avg:.5f}")
                    self.loss_df.loc[len(self.loss_df)] = [epoch, loss_avg, loss_normal_avg, loss_abnormal_avg]
                    val_loss = self.__calc_patches_validation_loss(val_loader, epoch, self.feature_matching_medium)
                if epoch % 10 == 0 and epoch != 0:
                    self.__save_model("medium/epoch_{}".format(epoch), self.feature_matching_medium)
                    self.loss_df.to_csv(loss_save_path, index=False)
                    self.val_loss_df.to_csv(val_loss_save_path, index=False)
                if epoch > 100:
                    if self.early_stopping and self.__stop_early(val_loss):
                        break
                if epoch == 100:
                    self.__set_new_lr(self.optimizer_medium)

                torch.cuda.empty_cache()

        # save model
        self.__save_model("medium/final", self.feature_matching_medium)

        self.loss_df.to_csv(loss_save_path, index=False)
        self.val_loss_df.to_csv(val_loss_save_path, index=False)

        self.__log_message("Matching Net for medium patches Trained.")

        self.best_val_loss = float("inf")
        self.epochs_without_improvement = 0

    def __train_small_patches(self):
        with task("dataset"):
            train_loader, val_loader = self.dataset.get_small_patches_train_and_val_dataloader(self.batch_size,
                                                                                               self.train_split)

        with task("train"):
            self.feature_extraction.eval()
            self.feature_matching_small.train()

            self.__log_message("Started training small patches...")
            loss_avg = 0.
            loss_normal_avg = 0.
            loss_abnormal_avg = 0.

            loss_save_path = os.path.join(self.output_dir, "small", "loss.csv")
            val_loss_save_path = os.path.join(self.output_dir, "small", "val_loss.csv")

            best_loss = float("inf")

            for epoch in range(self.epochs):
                self.feature_matching_small.train()

                for normals, abnormals, normal_masks, abnormal_masks in train_loader:  # for all train images
                    for normal, abnormal, normal_mask, abnormal_mask in zip(normals, abnormals, normal_masks,
                                                                            abnormal_masks):
                        normal = normal.to(self.device)
                        abnormal = abnormal.to(self.device)

                        self.optimizer_small.zero_grad()

                        with task('normal'):
                            surrogate_label_normal = self.feature_extraction(normal, self.feat_layers)
                            pred = self.feature_matching_small(normal, self.feat_layers)
                            loss_normal = 0
                            for feat_layer in self.feat_layers:
                                loss_normal += self.loss(pred[feat_layer], surrogate_label_normal[feat_layer])
                            loss_normal = loss_normal / len(self.feat_layers)

                        with task('abnormal'):
                            pred = self.feature_matching_small(abnormal, self.feat_layers)
                            loss_abnormal = 0
                            for feat_layer in self.feat_layers:
                                loss_abnormal += self.loss(pred[feat_layer], surrogate_label_normal[feat_layer])
                            loss_abnormal = loss_abnormal / len(self.feat_layers)

                        alpha = 1
                        loss = loss_normal + alpha * loss_abnormal
                        loss.backward()
                        self.optimizer_small.step()

                        # exponential moving average
                        loss_avg = loss_avg * 0.9 + float(loss.item()) * 0.1
                        loss_normal_avg = loss_normal_avg * 0.9 + float(loss_normal.item()) * 0.1
                        loss_abnormal = alpha * loss_abnormal
                        loss_abnormal_avg = loss_abnormal_avg * 0.9 + float(loss_abnormal.item()) * 0.1

                if loss_avg < best_loss:
                    best_loss = loss_avg
                    self.__save_model("small/best", self.feature_matching_small)
                if epoch % 1 == 0:
                    self.__log_message(f"Epoch {epoch}, loss = {loss_avg:.5f}, loss_normal = {loss_normal_avg:.5f}, "
                                       f"loss_abnormal = {loss_abnormal_avg:.5f}")
                    self.loss_df.loc[len(self.loss_df)] = [epoch, loss_avg, loss_normal_avg, loss_abnormal_avg]
                    val_loss = self.__calc_patches_validation_loss(val_loader, epoch, self.feature_matching_small)
                if epoch % 10 == 0 and epoch != 0:
                    self.__save_model("small/epoch_{}".format(epoch), self.feature_matching_small)
                    self.loss_df.to_csv(loss_save_path, index=False)
                    self.val_loss_df.to_csv(val_loss_save_path, index=False)
                if epoch > 100:
                    if self.early_stopping and self.__stop_early(val_loss):
                        break
                if epoch == 100:
                    self.__set_new_lr(self.optimizer_small)

                torch.cuda.empty_cache()

        # save model
        self.__save_model("small/final", self.feature_matching_small)

        self.loss_df.to_csv(loss_save_path, index=False)
        self.val_loss_df.to_csv(val_loss_save_path, index=False)

        self.__log_message("Matching Net for small patches Trained.")

        self.best_val_loss = float("inf")
        self.epochs_without_improvement = 0

    # endregion

    # region private methods

    def __calc_validation_loss(self, val_data_loader, epoch, model):
        model.eval()
        loss_avg = 0.
        loss_normal_avg = 0.
        loss_abnormal_avg = 0.

        for normal, abnormal, normal_mask, abnormal_mask in val_data_loader:
            normal = Resize(size=256, interpolation=InterpolationMode.BILINEAR)(normal).to(self.device)
            abnormal = Resize(size=256, interpolation=InterpolationMode.BILINEAR)(abnormal).to(self.device)

            with task('normal'):
                surrogate_label_normal = self.feature_extraction(normal, self.feat_layers)
                pred = model(normal, self.feat_layers)
                loss_normal = 0
                for feat_layer in self.feat_layers:
                    loss_normal += self.loss(pred[feat_layer], surrogate_label_normal[feat_layer])
                loss_normal = loss_normal / len(self.feat_layers)

            with task('abnormal'):
                pred = model(abnormal, self.feat_layers)
                loss_abnormal = 0
                for feat_layer in self.feat_layers:
                    loss_abnormal += self.loss(pred[feat_layer], surrogate_label_normal[feat_layer])
                loss_abnormal = loss_abnormal / len(self.feat_layers)

            alpha = 1
            loss = loss_normal + alpha * loss_abnormal

            # exponential moving average
            loss_avg = loss_avg * 0.9 + float(loss.item()) * 0.1
            loss_normal_avg = loss_normal_avg * 0.9 + float(loss_normal.item()) * 0.1
            loss_abnormal = alpha * loss_abnormal
            loss_abnormal_avg = loss_abnormal_avg * 0.9 + float(loss_abnormal.item()) * 0.1

        self.val_loss_df.loc[len(self.val_loss_df)] = [epoch, loss_avg, loss_normal_avg, loss_abnormal_avg]

        return loss_avg

    def __calc_patches_validation_loss(self, val_data_loader, epoch, model):
        model.eval()
        loss_avg = 0.
        loss_normal_avg = 0.
        loss_abnormal_avg = 0.

        for normals, abnormals, normal_masks, abnormal_masks in val_data_loader:  # for all train images
            for normal, abnormal, normal_mask, abnormal_mask in zip(normals, abnormals, normal_masks,
                                                                    abnormal_masks):
                normal = Resize(size=256, interpolation=InterpolationMode.BILINEAR)(normal).to(self.device)
                abnormal = Resize(size=256, interpolation=InterpolationMode.BILINEAR)(abnormal).to(self.device)

                with task('normal'):
                    surrogate_label_normal = self.feature_extraction(normal, self.feat_layers)
                    pred = model(normal, self.feat_layers)
                    loss_normal = 0
                    for feat_layer in self.feat_layers:
                        loss_normal += self.loss(pred[feat_layer], surrogate_label_normal[feat_layer])
                    loss_normal = loss_normal / len(self.feat_layers)

                with task('abnormal'):
                    pred = model(abnormal, self.feat_layers)
                    loss_abnormal = 0
                    for feat_layer in self.feat_layers:
                        loss_abnormal += self.loss(pred[feat_layer], surrogate_label_normal[feat_layer])
                    loss_abnormal = loss_abnormal / len(self.feat_layers)

                alpha = 1
                loss = loss_normal + alpha * loss_abnormal

                # exponential moving average
                loss_avg = loss_avg * 0.9 + float(loss.item()) * 0.1
                loss_normal_avg = loss_normal_avg * 0.9 + float(loss_normal.item()) * 0.1
                loss_abnormal = alpha * loss_abnormal
                loss_abnormal_avg = loss_abnormal_avg * 0.9 + float(loss_abnormal.item()) * 0.1

        self.val_loss_df.loc[len(self.val_loss_df)] = [epoch, loss_avg, loss_normal_avg, loss_abnormal_avg]

        return loss_avg

    def __stop_early(self, loss) -> bool:
        if loss < self.best_val_loss:
            self.best_val_loss = loss
            self.epochs_without_improvement = 0

            return False

        self.epochs_without_improvement += 1

        return self.epochs_without_improvement > self.patience

    def __set_new_lr(self, optimizer):
        for g in optimizer.param_groups:
            g['lr'] = 2e-5

    def __save_model(self, name: str, model):
        save_dir = os.path.join(self.output_dir, name)
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)

        save_path = os.path.join(save_dir, 'match.pt')
        torch.save(model.state_dict(), save_path)

    def __log_message(self, message):
        if self.debugging:
            print(message)

    # endregion



import numpy as np
import torch
import torch.nn.functional as F
import torchvision.transforms.functional as TF
import torchvision
from abc import ABC
from collections import OrderedDict
from os.path import join
from typing import List
from PIL import Image
from torch import Tensor
from torchvision.models import wide_resnet50_2, Wide_ResNet50_2_Weights
from torchvision.transforms import transforms

from src.anoseg.models.utils import BaseTester

"""
    Implementation of Sub-Image Anomaly Detection with Deep Pyramid Correspondences
    Code modified from https://github.com/byungjae89/SPADE-pytorch/tree/master
    Paper: https://arxiv.org/abs/2005.02357
"""


class Tester(BaseTester, ABC):

    # region init

    def __init__(self, model_path: str, debugging: bool = False,
                 image_size: int = 256, mask_size: int = 1024, use_self_ensembling: bool = False,
                 rot_90: bool = False, rot_180: bool = False, rot_270: bool = False, h_flip: bool = False,
                 h_flip_rot_90: bool = False, h_flip_rot_180: bool = False, h_flip_rot_270: bool = False,
                 integration_limit: float = 0.3, top_k: int = 5):
        self.device = 'cuda' if torch.cuda.is_available() else 'cpu'

        super().__init__(model_path=model_path, debugging=debugging, image_size=image_size, mask_size=mask_size,
                         use_self_ensembling=use_self_ensembling, rot_90=rot_90, rot_180=rot_180, rot_270=rot_270,
                         h_flip=h_flip, h_flip_rot_90=h_flip_rot_90, h_flip_rot_180=h_flip_rot_180,
                         h_flip_rot_270=h_flip_rot_270, integration_limit=integration_limit)
        self.top_k = top_k

        self.__load_model()
        self.__register_hooks()

    def __register_hooks(self):
        self.outputs = []

        def hook(module, input, output):
            self.outputs.append(output)

        self.model.layer1[-1].register_forward_hook(hook)
        self.model.layer2[-1].register_forward_hook(hook)
        self.model.layer3[-1].register_forward_hook(hook)
        self.model.avgpool.register_forward_hook(hook)

    def __load_model(self):
        l1 = torch.from_numpy(np.load(join(self.model_path, "layer_1.npy"))).to(self.device)
        l2 = torch.from_numpy(np.load(join(self.model_path, "layer_2.npy"))).to(self.device)
        l3 = torch.from_numpy(np.load(join(self.model_path, "layer_3.npy"))).to(self.device)
        pool = torch.from_numpy(np.load(join(self.model_path, "avgpool.npy"))).to(self.device)

        self.train_outputs = [l1, l2, l3, pool]

        self.model = wide_resnet50_2(weights=Wide_ResNet50_2_Weights.IMAGENET1K_V1, progress=True)
        self.model.to(self.device)
        self.model.eval()

    # endregion

    # region implement abstract methods

    def score(self, img_input) -> np.array:
        test_outputs = OrderedDict([('layer1', []), ('layer2', []), ('layer3', []), ('avgpool', [])])
        test_outputs_list = [[], [], [], []]

        with torch.no_grad():
            self.model(img_input.to(self.device))

        for k, v in zip(test_outputs.keys(), self.outputs):
            test_outputs[k].append(v)
        test_outputs_list = [lst + [self.outputs[i]] for i, lst in enumerate(test_outputs_list)]
        self.outputs.clear()

        for k, v in test_outputs.items():
            test_outputs[k] = torch.cat(v, 0)
        test_outputs_list = [torch.stack(lst, dim=0) for lst in test_outputs_list]

        dist_matrix = self.__calc_dist_matrix(torch.flatten(test_outputs_list[3], 1),
                                              torch.flatten(self.train_outputs[3], 1))

        # select K nearest neighbor and take average
        topk_values, topk_indexes = torch.topk(dist_matrix, k=self.top_k, dim=1, largest=False)
        # scores = torch.mean(topk_values, 1).cpu().detach().numpy()

        score_maps = []
        # for layer_name in ['layer1', 'layer2', 'layer3']:  # for each layer
        for i in range(3):

            # construct a gallery of features at all pixel locations of the K nearest neighbors
            topk_feat_map = self.train_outputs[i][topk_indexes[0]]
            test_feat_map = test_outputs_list[i][0]
            # original:
            """feat_gallery = topk_feat_map.transpose(3, 1).flatten(0, 2).unsqueeze(-1).unsqueeze(-1)"""

            # modified:
            # src: https://github.com/byungjae89/SPADE-pytorch/pull/18
            # adjust dimensions to measure distance in the channel dimension for all combinations
            feat_gallery = topk_feat_map.transpose(1, 2).transpose(2, 3)  # (K, C, H, W) -> (K, H, W, C)
            feat_gallery = feat_gallery.flatten(0, 2)  # (K, H, W, C) -> (KHW, C)
            feat_gallery = feat_gallery.unsqueeze(1).unsqueeze(1)  # (KHW, C) -> (KHW, 1, 1, C)
            test_feat_map = test_feat_map.transpose(1, 2).transpose(2, 3)  # (K, C, H, W) -> (K, H, W, C)

            # calculate distance matrix
            dist_matrix_list = []
            # original:
            """for d_idx in range(feat_gallery.shape[0] // 100):"""
            # modified:
            for d_idx in range(feat_gallery.shape[0] // 100 + 1):
                dist_matrix = torch.pairwise_distance(feat_gallery[d_idx * 100:d_idx * 100 + 100], test_feat_map)
                dist_matrix_list.append(dist_matrix.cpu())
            dist_matrix = torch.cat(dist_matrix_list, 0)

            # k nearest features from the gallery (k=1)
            score_map = torch.min(dist_matrix, dim=0)[0]

            score_map = F.interpolate(score_map.unsqueeze(0).unsqueeze(0), size=self.mask_size,
                                      mode='bilinear', align_corners=False)
            score_maps.append(score_map)

        # average distance between the features
        score_map = torch.mean(torch.cat(score_maps, 0), dim=0)

        # apply gaussian smoothing on the score map
        # score_map = gaussian_filter(score_map.squeeze().cpu().detach().numpy(), sigma=4)
        score_map = score_map.squeeze().cpu().detach().numpy()

        return score_map

    def score_with_augmentation(self, img_input):
        score_list = self._get_self_ensembling_scores(img_input)
        final_score = self._combine_scores(score_list)

        return final_score

    def preprocess_img(self, image_path: str, mean: List[float], std: List[float]) -> Tensor:
        original = Image.open(image_path).convert('RGB')

        normalize = transforms.Normalize(mean=mean, std=std)
        resize = torchvision.transforms.Resize(size=self.image_size, interpolation=TF.InterpolationMode.BILINEAR,
                                               antialias=True)

        transform = transforms.Compose([transforms.ToTensor(), normalize, resize])
        preprocessed = transform(original)

        preprocessed = preprocessed[None, :]

        return preprocessed

    # endregion

    # region private methods

    def __calc_dist_matrix(self, x, y):
        """Calculate Euclidean distance matrix with torch.tensor"""
        n = x.size(0)
        m = y.size(0)
        d = x.size(1)
        x = x.unsqueeze(1).expand(n, m, d)
        y = y.unsqueeze(0).expand(n, m, d)
        dist_matrix = torch.sqrt(torch.pow(x - y, 2).sum(2))

        return dist_matrix

    # endregion



import os
import numpy as np
import torch
from PIL import Image
from typing import Tuple
from typing_extensions import Literal

from src.anoseg.datasets.dataset import Dataset
from src.anoseg.utils import visualization
from src.anoseg.models.DFC.base_dfc.tester import Tester
from src.anoseg.models.DFC.patch_dfc.tester import Tester as PatchTester
from src.anoseg.models.DFC.base_dfc.trainer import Trainer
from src.anoseg.models.DFC.patch_dfc.trainer import Trainer as PatchTrainer

"""
    Implementation of DFC: Learning deep feature correspondence for unsupervised anomaly detection and segmentation
    Base implementation: https://github.com/YoungGod/DFC
    Paper: https://www.sciencedirect.com/science/article/abs/pii/S0031320322003557
"""


class DFC(object):

    # region init

    _DATASET_TYPES = Literal["textures", "objects"]

    def __init__(self, model_path: str = None):
        self.device = 'cuda' if torch.cuda.is_available() else 'cpu'
        self.use_patches = None

        if model_path is None:
            self.trained = False
        else:
            self.valid_model = self.__check_model(model_path)
            self.trained = True
            self.model_path = model_path

    def __check_model(self, model_path: str) -> bool:
        if self.__valid_model(model_path):
            self.use_patches = False
            return True
        elif self.__valid_patch_model(model_path):
            self.use_patches = True
            return True

        return False

    # endregion

    # region public methods

    def train(self, dataset: Dataset, dataset_type: _DATASET_TYPES, output_dir: str, use_patches: bool = False,
              debugging: bool = False, batch_size: int = 8, epochs: int = 301, lr: float = 2e-4,
              train_split: float = 0.9, pretrained_weights_dir: str = None, imagenet_dir: str = None,
              early_stopping: bool = True, patience: int = 10) -> None:
        self.use_patches = use_patches

        if self.use_patches:
            trainer = PatchTrainer(output_dir=output_dir, dataset=dataset, dataset_type=dataset_type,
                                   batch_size=batch_size, n_epochs=epochs, lr=lr, train_split=train_split,
                                   pretrained_weights_dir=pretrained_weights_dir, imagenet_dir=imagenet_dir,
                                   early_stopping=early_stopping, patience=patience, debugging=debugging)
        else:
            trainer = Trainer(output_dir=output_dir, dataset=dataset, dataset_type=dataset_type, batch_size=batch_size,
                              n_epochs=epochs, lr=lr, train_split=train_split,
                              pretrained_weights_dir=pretrained_weights_dir, imagenet_dir=imagenet_dir,
                              early_stopping=early_stopping, patience=patience, debugging=debugging)

        trainer.train()

    def eval(self, dataset: Dataset, dataset_type: _DATASET_TYPES, pretrained_weights_dir: str = None,
             debugging: bool = False, self_ensembling: bool = False,
             image_size: int = 256, mask_size: int = 1024,
             rot_90: bool = False, rot_180: bool = False, rot_270: bool = False,
             h_flip: bool = False, h_flip_rot_90: bool = False,
             h_flip_rot_180: bool = False, h_flip_rot_270: bool = False,
             integration_limit: float = 0.3):
        if not self.trained:
            raise Exception("Model not trained.")
        if not self.valid_model:
            raise Exception("Invalid model.")

        if self.use_patches:
            tester = PatchTester(model_path=self.model_path, debugging=debugging, dataset_type=dataset_type,
                                 pretrained_weights_dir=pretrained_weights_dir,
                                 image_size=image_size, mask_size=mask_size, use_self_ensembling=self_ensembling,
                                 rot_90=rot_90, rot_180=rot_180, rot_270=rot_270, h_flip=h_flip,
                                 h_flip_rot_90=h_flip_rot_90, h_flip_rot_180=h_flip_rot_180,
                                 h_flip_rot_270=h_flip_rot_270,
                                 integration_limit=integration_limit)
        else:
            tester = Tester(model_path=self.model_path, debugging=debugging, dataset_type=dataset_type,
                            pretrained_weights_dir=pretrained_weights_dir,
                            image_size=image_size, mask_size=mask_size, use_self_ensembling=self_ensembling,
                            rot_90=rot_90, rot_180=rot_180, rot_270=rot_270, h_flip=h_flip,
                            h_flip_rot_90=h_flip_rot_90, h_flip_rot_180=h_flip_rot_180,
                            h_flip_rot_270=h_flip_rot_270,
                            integration_limit=integration_limit)

        tester.evaluate(dataset=dataset)

    def display_predictions(self, dataset: Dataset, dataset_type: _DATASET_TYPES, pretrained_weights_dir: str = None,
                            debugging: bool = False, self_ensembling: bool = False,
                            image_size: int = 256, mask_size: int = 1024,
                            rot_90: bool = False, rot_180: bool = False, rot_270: bool = False,
                            h_flip: bool = False, h_flip_rot_90: bool = False,
                            h_flip_rot_180: bool = False, h_flip_rot_270: bool = False) -> None:
        if not self.trained:
            raise Exception("Model not trained.")
        if not self.valid_model:
            raise Exception("Invalid model.")

        if self.use_patches:
            tester = PatchTester(model_path=self.model_path, dataset_type=dataset_type,
                                 pretrained_weights_dir=pretrained_weights_dir, debugging=debugging,
                                 image_size=image_size, mask_size=mask_size, use_self_ensembling=self_ensembling,
                                 rot_90=rot_90, rot_180=rot_180, rot_270=rot_270, h_flip=h_flip,
                                 h_flip_rot_90=h_flip_rot_90, h_flip_rot_180=h_flip_rot_180,
                                 h_flip_rot_270=h_flip_rot_270)
        else:
            tester = Tester(model_path=self.model_path, dataset_type=dataset_type,
                            pretrained_weights_dir=pretrained_weights_dir, debugging=debugging,
                            image_size=image_size, mask_size=mask_size, use_self_ensembling=self_ensembling,
                            rot_90=rot_90, rot_180=rot_180, rot_270=rot_270, h_flip=h_flip,
                            h_flip_rot_90=h_flip_rot_90, h_flip_rot_180=h_flip_rot_180,
                            h_flip_rot_270=h_flip_rot_270)

        tester.display_predictions(dataset=dataset)

    def predict(self, image_path: str, display_prediction: bool, dataset_type: _DATASET_TYPES,
                pretrained_weights_dir: str = None, debugging: bool = False,
                mean: Tuple[float] = (0.485, 0.456, 0.406), std: Tuple[float] = (0.229, 0.224, 0.225),
                self_ensembling: bool = False, image_size: int = 256,
                rot_90: bool = False, rot_180: bool = False, rot_270: bool = False,
                h_flip: bool = False, h_flip_rot_90: bool = False,
                h_flip_rot_180: bool = False, h_flip_rot_270: bool = False) -> Tuple[np.array, np.array]:
        if not self.trained:
            raise Exception("Model not trained.")
        if not self.valid_model:
            raise Exception("Invalid model.")

        if self.use_patches:
            tester = PatchTester(model_path=self.model_path, dataset_type=dataset_type,
                                 pretrained_weights_dir=pretrained_weights_dir, debugging=debugging,
                                 image_size=image_size, use_self_ensembling=self_ensembling,
                                 rot_90=rot_90, rot_180=rot_180, rot_270=rot_270, h_flip=h_flip,
                                 h_flip_rot_90=h_flip_rot_90, h_flip_rot_180=h_flip_rot_180,
                                 h_flip_rot_270=h_flip_rot_270)
        else:
            tester = Tester(model_path=self.model_path, dataset_type=dataset_type,
                            pretrained_weights_dir=pretrained_weights_dir, debugging=debugging,
                            image_size=image_size, use_self_ensembling=self_ensembling,
                            rot_90=rot_90, rot_180=rot_180, rot_270=rot_270, h_flip=h_flip,
                            h_flip_rot_90=h_flip_rot_90, h_flip_rot_180=h_flip_rot_180,
                            h_flip_rot_270=h_flip_rot_270)

        score, binary_score = tester.predict(image_path=image_path, mean=mean, std=std)

        if display_prediction:
            original = Image.open(image_path).convert('RGB')

            visualization.display_images(img_list=[original, score, binary_score],
                                         titles=['original', 'score', 'binary_score'],
                                         cols=3)

        return score, binary_score

    # endregion

    # region private methods

    def __valid_model(self, model_path: str) -> bool:
        valid = os.path.exists(os.path.join(model_path, "match.pt"))

        return valid

    def __valid_patch_model(self, model_path: str) -> bool:
        valid = True
        valid = valid and self.__valid_model(os.path.join(model_path, "big"))
        valid = valid and self.__valid_model(os.path.join(model_path, "medium"))
        valid = valid and self.__valid_model(os.path.join(model_path, "small"))

        return valid

    # endregion



"""
Utils module

The code from this file comes from:
    * https://github.com/taikiinoue45/PaDiM
"""
import torch
import numpy as np
import pandas as pd
import torch.nn.functional as F

from numpy import ndarray as NDArray
from skimage import measure
from sklearn.metrics import auc, roc_auc_score, roc_curve
from torch import Tensor


def embeddings_concat(x0: Tensor, x1: Tensor) -> Tensor:
    b0, c0, h0, w0 = x0.size()
    _, c1, h1, w1 = x1.size()
    s = h0 // h1
    x0 = F.unfold(x0, kernel_size=(s, s), dilation=(1, 1), stride=(s, s))
    x0 = x0.view(b0, c0, -1, h1, w1)
    z = torch.zeros(b0, c0 + c1, x0.size(2), h1, w1).to(x0.device)
    for i in range(x0.size(2)):
        z[:, :, i, :, :] = torch.cat((x0[:, :, i, :, :], x1), 1)
    z = z.view(b0, -1, h1 * w1)
    z = F.fold(z, kernel_size=(s, s), output_size=(h0, w0), stride=(s, s))
    return z


def mean_smoothing(amaps: Tensor, kernel_size: int = 21) -> Tensor:
    mean_kernel = torch.ones(1, 1, kernel_size, kernel_size) / kernel_size ** 2
    mean_kernel = mean_kernel.to(amaps.device)
    return F.conv2d(amaps, mean_kernel, padding=kernel_size // 2, groups=1)


def compute_roc_score(predictions, ground_truth) -> float:
    num_data = len(predictions)

    amaps = np.stack(predictions)
    y_trues = np.stack(ground_truth)

    y_scores = amaps.max(1).max(1)
    y_trues = y_trues.any(axis=1).any(axis=1)

    # Save roc_curve.csv
    # keys = [f"threshold_{i}" for i in range(len(thresholds))]
    # roc_df = pd.DataFrame({"key": keys, "fpr": fprs, "tpr": tprs, "threshold": thresholds})
    # roc_df.to_csv("roc_curve.csv", index=False)

    # Update test_dataset.csv
    # pred_csv = pd.merge(
    #     pd.DataFrame({"stem": stems, "y_score": y_scores, "y_true": y_trues}),
    #     pd.read_csv("test_dataset.csv"),
    #     on="stem",
    # )
    # for i, th in enumerate(thresholds):
    #     pred_csv[f"threshold_{i}"] = pred_csv["y_score"].apply(lambda x: 1 if x >= th else 0)
    # pred_csv.to_csv("test_dataset.csv", index=False)



    return roc_auc_score(y_trues, y_scores)


def compute_pro_score(amaps: NDArray, masks: NDArray) -> float:

    df = pd.DataFrame([], columns=["pro", "fpr", "threshold"])
    binary_amaps = np.zeros_like(amaps, dtype=np.bool)

    max_step = 200
    min_th = amaps.min()
    max_th = amaps.max()
    delta = (max_th - min_th) / max_step

    for th in tqdm(np.arange(min_th, max_th, delta), desc="compute pro"):
        binary_amaps[amaps <= th] = 0
        binary_amaps[amaps > th] = 1

        pros = []
        for binary_amap, mask in zip(binary_amaps, masks):
            for region in measure.regionprops(measure.label(mask)):
                axes0_ids = region.coords[:, 0]
                axes1_ids = region.coords[:, 1]
                TP_pixels = binary_amap[axes0_ids, axes1_ids].sum()
                pros.append(TP_pixels / region.area)

        inverse_masks = 1 - masks
        FP_pixels = np.logical_and(inverse_masks, binary_amaps).sum()
        fpr = FP_pixels / inverse_masks.sum()

        df = df.append({"pro": mean(pros), "fpr": fpr, "threshold": th}, ignore_index=True)

    df.to_csv("pro_curve.csv", index=False)
    return auc(df["fpr"], df["pro"])


def denormalize(img: NDArray) -> NDArray:

    mean = np.array([0.485, 0.456, 0.406])
    std = np.array([0.229, 0.224, 0.225])
    img = (img * std + mean) * 255.0
    return img.astype(np.uint8)



import os
from typing import List, Tuple
from torch.utils.data import DataLoader

from src.anoseg.datasets.test_dataset import TestDataset
from src.anoseg.datasets.train_dataset import TrainDataset
from src.anoseg.datasets.patch_train_dataset import PatchTrainDataset


class Dataset(object):

    # region init

    def __init__(self, path_to_dataset: str, img_size: int = 256, mask_size: int = 1024, imagenet_dir: str = None,
                 self_supervised_training: bool = False,
                 mean: Tuple[float] = (0.485, 0.456, 0.406), std: Tuple[float] = (0.229, 0.224, 0.225),
                 anomaly_creation_method='dfc', dfc_anomaly_size='big', cutpaste_mode='all',
                 rot_90: bool = False, rot_180: bool = False, rot_270: bool = False, h_flip: bool = False,
                 h_flip_rot_90: bool = False, h_flip_rot_180: bool = False, h_flip_rot_270: bool = False):
        self.path_to_dataset = path_to_dataset
        self.img_size = img_size
        self.mask_size = mask_size
        self.imagenet_dir = imagenet_dir
        self.self_supervised_training = self_supervised_training
        self.mean = mean
        self.std = std

        self.anomaly_creation_method = anomaly_creation_method
        self.dfc_anomaly_size = dfc_anomaly_size
        self.cutpaste_mode = cutpaste_mode

        self.rot_90 = rot_90
        self.rot_180 = rot_180
        self.rot_270 = rot_270
        self.h_flip = h_flip
        self.h_flip_rot_90 = h_flip_rot_90
        self.h_flip_rot_180 = h_flip_rot_180
        self.h_flip_rot_270 = h_flip_rot_270

    # endregion

    # region public methods

    def get_train_dataloader(self, batch_size: int) -> DataLoader:
        image_paths = self.__get_train_img_paths()
        train_data = TrainDataset(image_paths=image_paths, imagenet_dir=self.imagenet_dir, img_size=self.img_size,
                                  mask_size=self.mask_size, rot_90=self.rot_90, rot_180=self.rot_180,
                                  rot_270=self.rot_270, h_flip=self.h_flip, h_flip_rot_90=self.h_flip_rot_90,
                                  h_flip_rot_180=self.h_flip_rot_180, h_flip_rot_270=self.h_flip_rot_270,
                                  self_supervised_training=self.self_supervised_training, mean=self.mean,
                                  std=self.std, dfc_anomaly_size=self.dfc_anomaly_size,
                                  method=self.anomaly_creation_method, cutpaste_mode=self.cutpaste_mode)
        train_loader = DataLoader(dataset=train_data, batch_size=batch_size, shuffle=True)

        return train_loader

    def get_train_and_val_dataloader(self, batch_size: int, train_split: float) -> Tuple[DataLoader, DataLoader]:
        image_paths = self.__get_train_img_paths()

        train_image_paths, val_image_paths = self.__split_image_paths(image_paths, train_split)

        train_data = TrainDataset(image_paths=train_image_paths, imagenet_dir=self.imagenet_dir, img_size=self.img_size,
                                  mask_size=self.mask_size, rot_90=self.rot_90, rot_180=self.rot_180,
                                  rot_270=self.rot_270, h_flip=self.h_flip, h_flip_rot_90=self.h_flip_rot_90,
                                  h_flip_rot_180=self.h_flip_rot_180, h_flip_rot_270=self.h_flip_rot_270,
                                  self_supervised_training=self.self_supervised_training, mean=self.mean,
                                  std=self.std, dfc_anomaly_size=self.dfc_anomaly_size,
                                  method=self.anomaly_creation_method, cutpaste_mode=self.cutpaste_mode)
        train_loader = DataLoader(dataset=train_data, batch_size=batch_size, shuffle=True)

        val_data = TrainDataset(image_paths=val_image_paths, imagenet_dir=self.imagenet_dir, img_size=self.img_size,
                                mask_size=self.mask_size, rot_90=False, rot_180=False,
                                rot_270=False, h_flip=False, h_flip_rot_90=False,
                                h_flip_rot_180=False, h_flip_rot_270=False,
                                self_supervised_training=self.self_supervised_training, mean=self.mean,
                                std=self.std, dfc_anomaly_size=self.dfc_anomaly_size,
                                method=self.anomaly_creation_method, cutpaste_mode=self.cutpaste_mode)
        val_loader = DataLoader(dataset=val_data, batch_size=1, shuffle=False)

        return train_loader, val_loader

    def get_medium_patches_train_dataloader(self, batch_size: int) -> DataLoader:
        return self.__get_patches_train_dataloader(batch_size=batch_size, img_size=self.img_size * 2)

    def get_medium_patches_train_and_val_dataloader(self, batch_size: int, train_split: float) \
            -> Tuple[DataLoader, DataLoader]:
        return self.__get_patches_train_and_val_dataloader(batch_size=batch_size, img_size=self.img_size * 2,
                                                           train_split=train_split)

    def get_small_patches_train_dataloader(self, batch_size: int) -> DataLoader:
        return self.__get_patches_train_dataloader(batch_size=batch_size, img_size=self.img_size * 4)

    def get_small_patches_train_and_val_dataloader(self, batch_size: int, train_split: float) \
            -> Tuple[DataLoader, DataLoader]:
        return self.__get_patches_train_and_val_dataloader(batch_size=batch_size, img_size=self.img_size * 4,
                                                           train_split=train_split)

    def get_test_dataloader(self) -> DataLoader:
        test_dataset = TestDataset(path_to_dataset=self.path_to_dataset, mask_size=self.mask_size,
                                   image_size=self.img_size, mean=self.mean, std=self.std)
        test_loader = DataLoader(dataset=test_dataset, batch_size=1, shuffle=False)

        return test_loader

    # endregion

    # region private methods

    def __get_patches_train_dataloader(self, batch_size: int, img_size: int) -> DataLoader:
        image_paths = self.__get_train_img_paths()
        patch_train_data = PatchTrainDataset(image_paths=image_paths, imagenet_dir=self.imagenet_dir,
                                             patch_size=self.img_size, img_size=img_size,
                                             mask_size=self.mask_size, rot_90=self.rot_90, rot_180=self.rot_180,
                                             rot_270=self.rot_270, h_flip=self.h_flip, h_flip_rot_90=self.h_flip_rot_90,
                                             h_flip_rot_180=self.h_flip_rot_180, h_flip_rot_270=self.h_flip_rot_270,
                                             self_supervised_training=self.self_supervised_training,
                                             mean=self.mean, std=self.std, dfc_anomaly_size=self.dfc_anomaly_size,
                                             method=self.anomaly_creation_method, cutpaste_mode=self.cutpaste_mode)
        train_loader = DataLoader(dataset=patch_train_data, batch_size=batch_size, shuffle=True)

        return train_loader

    def __get_patches_train_and_val_dataloader(self, batch_size: int, img_size: int, train_split: float) \
            -> Tuple[DataLoader, DataLoader]:
        image_paths = self.__get_train_img_paths()

        train_image_paths, val_image_paths = self.__split_image_paths(image_paths, train_split)

        patch_train_data = PatchTrainDataset(image_paths=train_image_paths, imagenet_dir=self.imagenet_dir,
                                             patch_size=self.img_size, img_size=img_size,
                                             mask_size=self.mask_size, rot_90=self.rot_90, rot_180=self.rot_180,
                                             rot_270=self.rot_270, h_flip=self.h_flip, h_flip_rot_90=self.h_flip_rot_90,
                                             h_flip_rot_180=self.h_flip_rot_180, h_flip_rot_270=self.h_flip_rot_270,
                                             self_supervised_training=self.self_supervised_training,
                                             mean=self.mean, std=self.std, dfc_anomaly_size=self.dfc_anomaly_size,
                                             method=self.anomaly_creation_method, cutpaste_mode=self.cutpaste_mode)
        train_loader = DataLoader(dataset=patch_train_data, batch_size=batch_size, shuffle=True)

        patch_val_data = PatchTrainDataset(image_paths=val_image_paths, imagenet_dir=self.imagenet_dir,
                                           patch_size=self.img_size, img_size=img_size,
                                           mask_size=self.mask_size, rot_90=False, rot_180=False,
                                           rot_270=False, h_flip=False, h_flip_rot_90=False,
                                           h_flip_rot_180=False, h_flip_rot_270=False,
                                           self_supervised_training=self.self_supervised_training,
                                           mean=self.mean, std=self.std, dfc_anomaly_size=self.dfc_anomaly_size,
                                           method=self.anomaly_creation_method, cutpaste_mode=self.cutpaste_mode)
        val_loader = DataLoader(dataset=patch_val_data, batch_size=1, shuffle=False)

        return train_loader, val_loader

    def __get_train_img_paths(self) -> List[str]:
        train_data_path = os.path.join(self.path_to_dataset, 'train/good')
        image_paths = []

        for root, dirs, files in os.walk(train_data_path):
            for file in files:
                image_paths.append(os.path.join(root, file))

        return image_paths

    def __split_image_paths(self, image_paths: List[str], train_split: float = 0.9) \
            -> Tuple[List[str], List[str]]:
        n = len(image_paths)
        n_train = int(n * train_split)

        train_image_paths = image_paths[:n_train]
        val_image_paths = image_paths[n_train:]

        return train_image_paths, val_image_paths

    # endregion



import torch
import torch.utils.data
import numpy as np
import os
import glob
from PIL import Image
import random
import torchvision.transforms as transforms


# src: https://github.com/YoungGod/DFC
# code was slightly adjusted

class ILSVRC(torch.utils.data.Dataset):
    def __init__(self, ilsvrc_data_path, mean, std, val=True):

        if val:
            synset_dir = os.path.join(ilsvrc_data_path, 'val')
        else:
            synset_dir = os.path.join(ilsvrc_data_path, 'train')
        self.img_paths = self.get_image_paths(synset_dir)

        normalize = transforms.Normalize(mean=mean,
                                         std=std)
        self.transform = transforms.Compose([transforms.ToTensor(), normalize])

    #         self.transform = transforms.Compose([transforms.RandomCrop(), transforms.ToTensor()])
    # self.transform = transform    # we define transforms outside the data set

    def __getitem__(self, index):
        path = self.img_paths[index]
        # print(sample)
        # read and transform
        img = self.preprocess_image(path)  # PIL image
        # print(img)
        # img=np.transpose(img,(2,0,1))
        # img = torch.from_numpy(img)
        img = np.array(img)
        img = self.transform(img)

        return img
        # return img    # only use the img

    def __len__(self):
        return len(self.img_paths)

    def preprocess_image(self, image_path):
        """ It reads an image, it resize it to have the lowest dimesnion of 256px,
            it randomly choose a 224x224 crop inside the resized image and normilize the numpy 
            array subtracting the ImageNet training set mean
            Args:
                images_path: path of the image
            Returns:
                cropped_im_array: the numpy array of the image normalized [width, height, channels]
        """
        # IMAGENET_MEAN = [123.68, 116.779, 103.939] # rgb format
        # print(image_path)
        img = Image.open(image_path).convert('RGB')
        # print(img)
        # resize of the image (setting lowest dimension to 256px)
        if img.size[0] < img.size[1]:
            h = int(float(256 * img.size[1]) / img.size[0])
            img = img.resize((256, h), Image.ANTIALIAS)  # 抗锯齿
        else:
            w = int(float(256 * img.size[0]) / img.size[1])
            img = img.resize((w, 256), Image.ANTIALIAS)

        # in case when the image size < (256, 256)
        if img.size[0] < 256 or img.size[1] < 256:
            img = img.resize((256, 256), Image.ANTIALIAS)

        # random 244x224 patch
        x = random.randint(0, img.size[0] - 224)
        y = random.randint(0, img.size[1] - 224)
        img_cropped = img.crop((x, y, x + 224, y + 224))

        # data augmentation: flip left right
        if random.randint(0, 1) == 1:
            img_cropped = img_cropped.transpose(Image.FLIP_LEFT_RIGHT)

        # cropped_im_array = np.array(img_cropped, dtype=np.float32)
        #
        # for i in range(3):
        #     cropped_im_array[:,:,i] -= IMAGENET_MEAN[i]
        #
        # return cropped_im_array/225
        return img_cropped

    def get_image_paths(self, data_path):
        img_paths = []
        pattern = data_path + "/*/*.JPEG"
        for path in glob.glob(pattern):
            img_paths.append(path)

        return img_paths



import torch
import os
import numpy as np

from PIL import Image
from typing import Tuple
from src.anoseg.datasets.dataset import Dataset
from src.anoseg.models.PaDiM.base_padim.trainer import Trainer
from src.anoseg.models.PaDiM.patch_padim.trainer import Trainer as PatchTrainer
from src.anoseg.models.PaDiM.base_padim.tester import Tester
from src.anoseg.models.PaDiM.patch_padim.tester import Tester as PatchTester
from src.anoseg.utils import visualization

"""
    Implementation of PaDiM: a Patch Distribution Modeling Framework for Anomaly Detection and Localization
    Base implementation: https://github.com/Pangoraw/PaDiM
    Paper: https://arxiv.org/abs/2011.08785
"""


class PaDiM(object):

    # region init

    def __init__(self, model_path: str = None):
        self.device = 'cuda' if torch.cuda.is_available() else 'cpu'
        self.use_patches = None

        if model_path is None:
            self.trained = False
        else:
            self.valid_model = self.__check_model(model_path)
            self.trained = True
            self.model_path = model_path

    def __check_model(self, model_path: str) -> bool:
        if self.__valid_model(model_path):
            self.use_patches = False
            return True
        elif self.__valid_patch_model(model_path):
            self.use_patches = True
            return True

        return False

    # endregion

    # region public methods

    def train(self, dataset, output_dir: str, use_patches: bool = False, debugging: bool = False,
              num_embeddings=130, backbone="resnet18", image_size: int = 512, batch_size: int = 30) -> None:
        self.use_patches = use_patches

        if self.use_patches:
            trainer = PatchTrainer(output_dir=output_dir,
                                   dataset=dataset,
                                   batch_size=batch_size,
                                   num_embeddings=num_embeddings,
                                   backbone=backbone,
                                   image_size=image_size,
                                   debugging=debugging)
        else:
            trainer = Trainer(output_dir=output_dir,
                              dataset=dataset,
                              batch_size=batch_size,
                              num_embeddings=num_embeddings,
                              backbone=backbone,
                              image_size=image_size,
                              debugging=debugging)

        trainer.train()

    def eval(self, dataset: Dataset, debugging: bool = False, self_ensembling: bool = False,
             image_size: int = 256, mask_size: int = 1024,
             rot_90: bool = False, rot_180: bool = False, rot_270: bool = False,
             h_flip: bool = False, h_flip_rot_90: bool = False,
             h_flip_rot_180: bool = False, h_flip_rot_270: bool = False,
             integration_limit: float = 0.3, backbone: str = 'wide_resnet50') -> None:
        if not self.trained:
            raise Exception("Model not trained.")
        if not self.valid_model:
            raise Exception("Invalid model.")

        if self.use_patches:
            tester = PatchTester(model_path=self.model_path, debugging=debugging,
                                 image_size=image_size, mask_size=mask_size, use_self_ensembling=self_ensembling,
                                 rot_90=rot_90, rot_180=rot_180, rot_270=rot_270, h_flip=h_flip,
                                 h_flip_rot_90=h_flip_rot_90, h_flip_rot_180=h_flip_rot_180,
                                 h_flip_rot_270=h_flip_rot_270,
                                 integration_limit=integration_limit, backbone=backbone)
        else:
            tester = Tester(model_path=self.model_path, debugging=debugging,
                            image_size=image_size, mask_size=mask_size, use_self_ensembling=self_ensembling,
                            rot_90=rot_90, rot_180=rot_180, rot_270=rot_270, h_flip=h_flip,
                            h_flip_rot_90=h_flip_rot_90, h_flip_rot_180=h_flip_rot_180,
                            h_flip_rot_270=h_flip_rot_270,
                            integration_limit=integration_limit, backbone=backbone)

        tester.evaluate(dataset=dataset)

    def display_predictions(self, dataset: Dataset, debugging: bool = False, self_ensembling: bool = False,
                            image_size: int = 256, mask_size: int = 1024,
                            rot_90: bool = False, rot_180: bool = False, rot_270: bool = False,
                            h_flip: bool = False, h_flip_rot_90: bool = False,
                            h_flip_rot_180: bool = False, h_flip_rot_270: bool = False,
                            backbone: str = 'wide_resnet50') -> None:
        if not self.trained:
            raise Exception("Model not trained.")
        if not self.valid_model:
            raise Exception("Invalid model.")

        if self.use_patches:
            tester = PatchTester(model_path=self.model_path, debugging=debugging,
                                 image_size=image_size, mask_size=mask_size, use_self_ensembling=self_ensembling,
                                 rot_90=rot_90, rot_180=rot_180, rot_270=rot_270, h_flip=h_flip,
                                 h_flip_rot_90=h_flip_rot_90, h_flip_rot_180=h_flip_rot_180,
                                 h_flip_rot_270=h_flip_rot_270, backbone=backbone)
        else:
            tester = Tester(model_path=self.model_path, debugging=debugging,
                            image_size=image_size, mask_size=mask_size, use_self_ensembling=self_ensembling,
                            rot_90=rot_90, rot_180=rot_180, rot_270=rot_270, h_flip=h_flip,
                            h_flip_rot_90=h_flip_rot_90, h_flip_rot_180=h_flip_rot_180,
                            h_flip_rot_270=h_flip_rot_270, backbone=backbone)

        tester.display_predictions(dataset=dataset)

    def predict(self, image_path: str, display_prediction: bool, debugging: bool = False,
                mean: Tuple[float] = (0.485, 0.456, 0.406), std: Tuple[float] = (0.229, 0.224, 0.225),
                self_ensembling: bool = False, image_size: int = 256,
                rot_90: bool = False, rot_180: bool = False, rot_270: bool = False,
                h_flip: bool = False, h_flip_rot_90: bool = False,
                h_flip_rot_180: bool = False, h_flip_rot_270: bool = False,
                backbone: str = 'wide_resnet50') -> Tuple[np.array, np.array]:
        if not self.trained:
            raise Exception("Model not trained.")
        if not self.valid_model:
            raise Exception("Invalid model.")

        if self.use_patches:
            tester = PatchTester(model_path=self.model_path, debugging=debugging,
                                 image_size=image_size, use_self_ensembling=self_ensembling,
                                 rot_90=rot_90, rot_180=rot_180, rot_270=rot_270, h_flip=h_flip,
                                 h_flip_rot_90=h_flip_rot_90, h_flip_rot_180=h_flip_rot_180,
                                 h_flip_rot_270=h_flip_rot_270, backbone=backbone)
        else:
            tester = Tester(model_path=self.model_path, debugging=debugging,
                            image_size=image_size, use_self_ensembling=self_ensembling,
                            rot_90=rot_90, rot_180=rot_180, rot_270=rot_270, h_flip=h_flip,
                            h_flip_rot_90=h_flip_rot_90, h_flip_rot_180=h_flip_rot_180,
                            h_flip_rot_270=h_flip_rot_270, backbone=backbone)

        score, binary_score = tester.predict(image_path=image_path, mean=mean, std=std)

        if display_prediction:
            original = Image.open(image_path).convert('RGB')

            visualization.display_images(img_list=[original, score, binary_score],
                                         titles=['original', 'score', 'binary_score'],
                                         cols=3)

        return score, binary_score

    # endregion

    # region private methods

    def __valid_model(self, model_path: str) -> bool:
        valid = True
        valid = valid and os.path.exists(os.path.join(model_path, "covs.npy"))
        valid = valid and os.path.exists(os.path.join(model_path, "embedding_ids.npy"))
        valid = valid and os.path.exists(os.path.join(model_path, "means.npy"))
        valid = valid and os.path.exists(os.path.join(model_path, "n.npy"))

        return valid

    def __valid_patch_model(self, model_path: str) -> bool:
        valid = True
        valid = valid and self.__valid_model(os.path.join(model_path, "big"))
        valid = valid and self.__valid_model(os.path.join(model_path, "medium"))
        valid = valid and self.__valid_model(os.path.join(model_path, "small"))

        return valid

    # endregion



import numpy as np


def calculate_avg_iou(ground_truth, binary_scores):
    ious = []

    for binary_score, gt_score in zip(binary_scores, ground_truth):
        if gt_score.any() > 0:  # when the gt have no anomaly pixels, skip it
            iou = calculate_iou(binary_score, gt_score)
            ious.append(iou)

    return np.array(ious).mean()


def calculate_iou(prediction, mask):
    intersection = np.logical_and(prediction, mask).astype(np.float32).sum()
    union = np.logical_or(prediction, mask).astype(np.float32).sum()
    iou = intersection / union

    return iou
